import logging
from typing import List
import grpc
import time
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.host_managers_info import HostManagersInfo
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
import csle_collector.host_manager.host_manager_pb2_grpc
import csle_collector.host_manager.host_manager_pb2
import csle_collector.host_manager.query_host_manager
import csle_collector.host_manager.host_manager_util
from csle_common.util.emulation_util import EmulationUtil
from csle_common.logging.log import Logger


class HostController:
    """
    Class controlling host managers and host specific configurations
    """

    @staticmethod
    def start_host_managers(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method for checking if the host manager is running and starting it if it is not running

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """

        # Start host managers on emulation containers
        for c in emulation_env_config.containers_config.containers:
            # Connect
            HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=c.docker_gw_bridge_ip,
                                              logger=logger)

        # Start host manager on kafka container
        HostController.start_host_manager(emulation_env_config=emulation_env_config,
                                          ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                          logger=logger)

        # Start host manager on ELK container
        HostController.start_host_manager(emulation_env_config=emulation_env_config,
                                          ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                          logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            # Start host manager on SDN controller container
            HostController.start_host_manager(
                emulation_env_config=emulation_env_config,
                ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_host_manager(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        Utility method for starting the host manager on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if host_manager is already running
        status = None
        not_running = False
        try:
            status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                ip=ip, port=emulation_env_config.host_manager_config.host_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = (f"Monitor running: {status.monitor_running}, filebeat running: {status.filebeat_running},"
                          f" packetbeat running: {status.packetbeat_running}, "
                          f"metricbeat running: {status.metricbeat_running}, "
                          f"heartbeat running: {status.heartbeat_running}")

        # Extract container name for logging
        container = emulation_env_config.containers_config.get_container_from_ip(ip)
        alt_str = ""
        if container is not None:
            alt_str = f"{container.get_ips()[0]}, {container.get_full_name()}"

        if not_running:
            logger.info(f"Host manager is not running on: {ip} ({alt_str}), starting it.")
            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.HOST_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                    conn=emulation_env_config.get_connection(ip=ip))

            # Start the host_manager
            cmd = constants.COMMANDS.START_HOST_MANAGER.format(
                emulation_env_config.host_manager_config.host_manager_port,
                emulation_env_config.host_manager_config.host_manager_log_dir,
                emulation_env_config.host_manager_config.host_manager_log_file,
                emulation_env_config.host_manager_config.host_manager_max_workers)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                    conn=emulation_env_config.get_connection(ip=ip))
            time.sleep(2)
        else:
            logger.info(f"Host manager is already running on: {ip} ({alt_str}). Status: {status_str}")

    @staticmethod
    def stop_host_managers(emulation_env_config: EmulationEnvConfig, physical_host_ip: str) -> None:
        """
        Utility method for stopping host managers

        :param emulation_env_config: the emulation env config
        :param physical_host_ip: the ip of the physical host
        :return: None
        """
        # Stop host manager on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_host_ip:
                HostController.stop_host_manager(emulation_env_config=emulation_env_config, ip=c.docker_gw_bridge_ip)

        # Stop host manager on Kafka container
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_host_ip:
            HostController.stop_host_manager(emulation_env_config=emulation_env_config,
                                             ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip)

        # Stop host manager on ELK container
        if emulation_env_config.elk_config.container.physical_host_ip == physical_host_ip:
            HostController.stop_host_manager(emulation_env_config=emulation_env_config,
                                             ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip)

        if emulation_env_config.sdn_controller_config is not None:
            # Stop host manager on the SDN controller container
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_host_ip:
                HostController.stop_host_manager(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip)

    @staticmethod
    def stop_host_manager(emulation_env_config: EmulationEnvConfig, ip: str) -> None:
        """
        Utility method for stopping the host manager on a specific container

        :param emulation_env_config: the emulation env config
        :param ip: the ip of the container
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config, ip=ip)

        Logger.__call__().get_logger().info(f"Stopping host manager on node {ip}")

        # Stop old background job if running
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.HOST_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                conn=emulation_env_config.get_connection(ip=ip))

    @staticmethod
    def start_host_monitor_threads(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                   logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Start host monitor on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.start_host_monitor_thread(emulation_env_config=emulation_env_config,
                                                     ip=c.docker_gw_bridge_ip, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start host monitor on the Kafka container
            HostController.start_host_monitor_thread(
                emulation_env_config=emulation_env_config,
                ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Start host monitor on the ELK container
            HostController.start_host_monitor_thread(
                emulation_env_config=emulation_env_config,
                ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip, logger=logger)

        if emulation_env_config.sdn_controller_config is not None and \
                emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Start host monitor on the SDN controller container
            HostController.start_host_monitor_thread(
                emulation_env_config=emulation_env_config,
                ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def start_filebeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger,
                        initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and filebeat

        :param emulation_env_config: the emulation env config
        :param initial_start: boolean indicating whether this method is called on emulation initialziation or not
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Start filebeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.start_filebeat(emulation_env_config=emulation_env_config,
                                          ips=[c.docker_gw_bridge_ip] + c.get_ips(),
                                          initial_start=initial_start, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start filebeat on the Kafka container
            HostController.start_filebeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.kafka_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.kafka_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Start filebeat on the ELK container
            HostController.start_filebeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.elk_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.elk_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Start filebeat on the SDN controller container
            HostController.start_filebeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.sdn_controller_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

    @staticmethod
    def start_packetbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger,
                          initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and packetbeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param initial_start: boolean indicating whether this method is called on emulation initialziation or not
        :param logger: the logger to use for logging
        :return: None
        """
        # Start packetbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.start_packetbeat(
                emulation_env_config=emulation_env_config,
                ips=[c.docker_gw_bridge_ip] + c.get_ips(),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start packetbeat on the Kafka container
            HostController.start_packetbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.kafka_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.kafka_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Start packetbeat on the ELK container
            HostController.start_packetbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.elk_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.elk_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Start packetbeat on the SDN controller container
            HostController.start_packetbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.sdn_controller_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

    @staticmethod
    def start_metricbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger,
                          initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and metricbeat

        :param emulation_env_config: the emulation env config
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Start packetbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.start_metricbeat(
                emulation_env_config=emulation_env_config,
                ips=[c.docker_gw_bridge_ip] + c.get_ips(),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start metricbeat on the Kafka container
            HostController.start_metricbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.kafka_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.kafka_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start metricbeat on the ELK container
            HostController.start_metricbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.elk_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.elk_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Start metricbeat on the SDN controller container
            HostController.start_metricbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.sdn_controller_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

    @staticmethod
    def start_heartbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger,
                         initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and heartbeat

        :param emulation_env_config: the emulation env config
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Start heartbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.start_heartbeat(
                emulation_env_config=emulation_env_config,
                ips=[c.docker_gw_bridge_ip] + c.get_ips(),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Start heartbeat on the Kafka container
            HostController.start_heartbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.kafka_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.kafka_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Start heartbeat on the ELK container
            HostController.start_heartbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.elk_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.elk_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Start heartbeat on the SDN controller container
            HostController.start_heartbeat(
                emulation_env_config=emulation_env_config,
                ips=([emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip] +
                     emulation_env_config.sdn_controller_config.container.get_ips()),
                initial_start=initial_start, logger=logger)

    @staticmethod
    def stop_filebeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                       logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to stop filebeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the physical host ip
        :param logger: the logger to use for logging
        :return: None
        """
        # Stop filebeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip:
                HostController.stop_filebeat(emulation_env_config=emulation_env_config, ip=c.docker_gw_bridge_ip,
                                             logger=logger)

        # Stop filebeat on the kafka container
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_filebeat(emulation_env_config=emulation_env_config,
                                         ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                         logger=logger)

        # Stop filebeat on the ELK container
        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_filebeat(emulation_env_config=emulation_env_config,
                                         ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                         logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
                # Stop filebeat on the SDN controller container
                HostController.stop_filebeat(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                    logger=logger)

    @staticmethod
    def stop_packetbeats(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                         physical_server_ip: str) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to stop packetbeat

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :param physical_server_ip: the of of the physical host
        :return: None
        """
        # Stop packetbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip:
                HostController.stop_packetbeat(emulation_env_config=emulation_env_config, ip=c.docker_gw_bridge_ip,
                                               logger=logger)

        # Stop packetbeat on the kafka container
        if physical_server_ip == emulation_env_config.kafka_config.container.physical_host_ip:
            HostController.stop_packetbeat(emulation_env_config=emulation_env_config,
                                           ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                           logger=logger)

        # Stop packetbeat on the ELK container
        if physical_server_ip == emulation_env_config.elk_config.container.physical_host_ip:
            HostController.stop_packetbeat(emulation_env_config=emulation_env_config,
                                           ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                           logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
                # Stop packetbeat on the SDN controller container
                HostController.stop_packetbeat(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                    logger=logger)

    @staticmethod
    def stop_metricbeats(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                         physical_server_ip: str) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to stop metricbeat

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :param physical_server_ip: the physical server ip
        :return: None
        """
        # Stop metricbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip:
                HostController.stop_metricbeat(emulation_env_config=emulation_env_config, ip=c.docker_gw_bridge_ip,
                                               logger=logger)

        # Stop metricbeat on the kafka container
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_metricbeat(emulation_env_config=emulation_env_config,
                                           ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                           logger=logger)

        # Stop metricbeat on the ELK container
        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_metricbeat(emulation_env_config=emulation_env_config,
                                           ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                           logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
                # Stop metricbeat on the SDN controller container
                HostController.stop_metricbeat(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_heartbeats(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                        physical_server_ip: str) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to stop heartbeat

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :param physical_server_ip: the physical server ip
        :return: None
        """
        # Stop heartbeat on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip:
                HostController.stop_heartbeat(emulation_env_config=emulation_env_config,
                                              ip=c.docker_gw_bridge_ip, logger=logger)

        # Stop heartbeat on the kafka container
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_heartbeat(emulation_env_config=emulation_env_config,
                                          ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                          logger=logger)

        # Stop heartbeat on the ELK container
        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            HostController.stop_heartbeat(emulation_env_config=emulation_env_config,
                                          ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                          logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
                # Stop heartbeat on the SDN controller container
                HostController.stop_heartbeat(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                    logger=logger)

    @staticmethod
    def config_filebeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                         logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to setup the configuration of filebeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Configure filebeat on the emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.config_filebeat(emulation_env_config=emulation_env_config, container=c, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Configure filebeat on the kafka container
            HostController.config_filebeat(emulation_env_config=emulation_env_config,
                                           container=emulation_env_config.kafka_config.container, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Configure filebeat on the ELK container
            HostController.config_filebeat(emulation_env_config=emulation_env_config,
                                           container=emulation_env_config.elk_config.container, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Configure filebeat on the SDN controller container
            HostController.config_filebeat(emulation_env_config=emulation_env_config,
                                           container=emulation_env_config.sdn_controller_config.container,
                                           logger=logger)

    @staticmethod
    def config_packetbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                           logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to setup the configuration of packetbeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Configure packetbeat on the emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.config_packetbeat(emulation_env_config=emulation_env_config, container=c, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Configure packetbeat on the kafka container
            HostController.config_packetbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.kafka_config.container, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Configure packetbeat on the ELK container
            HostController.config_packetbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.elk_config.container, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Configure packetbeat on the SDN controller container
            HostController.config_packetbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.sdn_controller_config.container,
                                             logger=logger)

    @staticmethod
    def config_metricbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                           logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to setup the configuration of metricbeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Configure metricbeat on the emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.config_metricbeat(emulation_env_config=emulation_env_config, container=c, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Configure metricbeat on the kafka container
            HostController.config_metricbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.kafka_config.container, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Configure metricbeat on the ELK container
            HostController.config_metricbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.elk_config.container, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Configure metricbeat on the SDN controller container
            HostController.config_metricbeat(emulation_env_config=emulation_env_config,
                                             container=emulation_env_config.sdn_controller_config.container,
                                             logger=logger)

    @staticmethod
    def config_heartbeats(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on every container
        to start the Host manager and to setup the configuration of heartbeat

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        # Configure heartbeat on the emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            HostController.config_heartbeat(emulation_env_config=emulation_env_config, container=c, logger=logger)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Configure heartbeat on the kafka container
            HostController.config_heartbeat(emulation_env_config=emulation_env_config,
                                            container=emulation_env_config.kafka_config.container, logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Configure heartbeat on the ELK container
            HostController.config_heartbeat(emulation_env_config=emulation_env_config,
                                            container=emulation_env_config.elk_config.container, logger=logger)

        if emulation_env_config.sdn_controller_config is not None \
                and emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
            # Configure heartbeat on the SDN controller container
            HostController.config_heartbeat(emulation_env_config=emulation_env_config,
                                            container=emulation_env_config.sdn_controller_config.container,
                                            logger=logger)

    @staticmethod
    def start_host_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and the monitor thread

        :param emulation_env_config: the emulation env config
        :param ip: IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        host_monitor_dto = HostController.get_host_monitor_thread_status_by_port_and_ip(
            ip=ip, port=emulation_env_config.host_manager_config.host_manager_port)
        if not host_monitor_dto.monitor_running:
            logger.info(f"Host monitor thread is not running on {ip}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ip}:{emulation_env_config.host_manager_config.host_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
                csle_collector.host_manager.query_host_manager.start_host_monitor(
                    stub=stub, kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    kafka_port=emulation_env_config.kafka_config.kafka_port,
                    time_step_len_seconds=emulation_env_config.kafka_config.time_step_len_seconds)

    @staticmethod
    def start_filebeat(emulation_env_config: EmulationEnvConfig, ips: List[str], logger: logging.Logger,
                       initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and filebeat

        :param emulation_env_config: the emulation env config
        :param ips: IP of the container
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        if initial_start:
            node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=ips)
            if node_beats_config is None or not node_beats_config.start_filebeat_automatically:
                return
        host_monitor_dto = HostController.get_host_monitor_thread_status_by_port_and_ip(
            ip=ips[0], port=emulation_env_config.host_manager_config.host_manager_port)
        if not host_monitor_dto.filebeat_running:
            logger.info(f"Filebeat is not running on {ips[0]}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
                csle_collector.host_manager.query_host_manager.start_filebeat(stub=stub)
        else:
            logger.info(f"Filebeat is already running on {ips[0]}.")

    @staticmethod
    def start_packetbeat(emulation_env_config: EmulationEnvConfig, ips: List[str],
                         logger: logging.Logger, initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and packetbeat

        :param emulation_env_config: the emulation env config
        :param ips: IP of the container
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        if initial_start:
            node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=ips)
            if node_beats_config is None or not node_beats_config.start_packetbeat_automatically:
                return
        host_monitor_dto = HostController.get_host_monitor_thread_status_by_port_and_ip(
            ip=ips[0], port=emulation_env_config.host_manager_config.host_manager_port)
        if not host_monitor_dto.packetbeat_running:
            logger.info(
                f"Packetbeat is not running on {ips[0]}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
                csle_collector.host_manager.query_host_manager.start_packetbeat(stub=stub)
        else:
            logger.info(f"Packetbeat is already running on {ips[0]}.")

    @staticmethod
    def start_metricbeat(emulation_env_config: EmulationEnvConfig, ips: List[str],
                         logger: logging.Logger, initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and metricbeat

        :param emulation_env_config: the emulation env config
        :param ips: IP of the container
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        if initial_start:
            node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=ips)
            if node_beats_config is None or not node_beats_config.start_metricbeat_automatically:
                return
        host_monitor_dto = HostController.get_host_monitor_thread_status_by_port_and_ip(
            ip=ips[0], port=emulation_env_config.host_manager_config.host_manager_port)
        if not host_monitor_dto.metricbeat_running:
            logger.info(
                f"Metricbeat is not running on {ips[0]}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
                csle_collector.host_manager.query_host_manager.start_metricbeat(stub=stub)
        else:
            logger.info(f"Metricbeat is already running on {ips[0]}.")

    @staticmethod
    def start_heartbeat(emulation_env_config: EmulationEnvConfig, ips: List[str],
                        logger: logging.Logger,
                        initial_start: bool = False) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and heartbeat

        :param emulation_env_config: the emulation env config
        :param ips: IPs of the container
        :param initial_start: boolean indicating whether this method is called on emulation initialization or not
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        if initial_start:
            node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=ips)
            if node_beats_config is None or not node_beats_config.start_heartbeat_automatically:
                return
        host_monitor_dto = HostController.get_host_monitor_thread_status_by_port_and_ip(
            ip=ips[0], port=emulation_env_config.host_manager_config.host_manager_port)
        if not host_monitor_dto.heartbeat_running:
            logger.info(
                f"Heartbeat is not running on {ips[0]}, starting it.")
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
                csle_collector.host_manager.query_host_manager.start_heartbeat(stub=stub)
        else:
            logger.info(f"Heartbeat is already running on {ips[0]}.")

    @staticmethod
    def start_sparks(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Utility function for starting Spark on compute nodes

        :param emulation_config: the emulation env configuration
        :param physical_server_ip: the ip of the phsyical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SPARK_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Starting Spark on IP: {c.docker_gw_bridge_ip} ({c.get_ips()[0]}, "
                                f"{c.get_full_name()})")
                    HostController.start_spark(emulation_env_config=emulation_env_config, ips=[c.docker_gw_bridge_ip],
                                               logger=logger)

    @staticmethod
    def stop_sparks(emulation_env_config: EmulationEnvConfig, physical_server_ip: str, logger: logging.Logger) -> None:
        """
        Utility function for stopping Spark on compute nodes

        :param emulation_config: the emulation env configuration
        :param physical_server_ip: the ip of the phsyical server
        :param logger: the logger to use for logging
        :return: None
        """
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip != physical_server_ip:
                continue
            for ids_image in constants.CONTAINER_IMAGES.SPARK_IMAGES:
                if ids_image in c.name:
                    logger.info(f"Stopping Spark on IP: {c.docker_gw_bridge_ip} ({c.get_ips()[0]}, "
                                f"{c.get_full_name()})")
                    HostController.stop_spark(emulation_env_config=emulation_env_config,
                                              ips=[c.docker_gw_bridge_ip], logger=logger)

    @staticmethod
    def start_spark(emulation_env_config: EmulationEnvConfig, ips: List[str], logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to start the Host manager and spark

        :param emulation_env_config: the emulation env config
        :param ips: IPs of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            csle_collector.host_manager.query_host_manager.start_spark(stub=stub)

    @staticmethod
    def stop_spark(emulation_env_config: EmulationEnvConfig, ips: List[str], logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific IP
        to stop spark

        :param emulation_env_config: the emulation env config
        :param ips: IPs of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ips[0], logger=logger)
        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ips[0]}:{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            csle_collector.host_manager.query_host_manager.stop_spark(stub=stub)

    @staticmethod
    def config_filebeat(emulation_env_config: EmulationEnvConfig, container: NodeContainerConfig,
                        logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container
        to setup the filebeat configuration

        :param emulation_env_config: the emulation env config
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=container.docker_gw_bridge_ip,
                                          logger=logger)
        node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=container.get_ips())
        if node_beats_config is None:
            return
        kafka_topics = list(map(lambda topic: topic.name, emulation_env_config.kafka_config.topics))

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Configuring filebeat on {container.docker_gw_bridge_ip}, {container.get_full_name()}, "
                        f"ips: {container.get_ips()}")
            csle_collector.host_manager.query_host_manager.config_filebeat(
                stub=stub, log_files_paths=node_beats_config.log_files_paths,
                kibana_ip=emulation_env_config.elk_config.container.get_ips()[0],
                kibana_port=emulation_env_config.elk_config.kibana_port,
                elastic_ip=emulation_env_config.elk_config.container.get_ips()[0],
                elastic_port=emulation_env_config.elk_config.elastic_port,
                num_elastic_shards=emulation_env_config.beats_config.num_elastic_shards, kafka_topics=kafka_topics,
                kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                kafka_port=emulation_env_config.kafka_config.kafka_port,
                filebeat_modules=node_beats_config.filebeat_modules,
                reload_enabled=emulation_env_config.beats_config.reload_enabled,
                kafka=node_beats_config.kafka_input)

    @staticmethod
    def config_packetbeat(emulation_env_config: EmulationEnvConfig, container: NodeContainerConfig,
                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container
        to setup the packetbeat configuration

        :param emulation_env_config: the emulation env config
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=container.docker_gw_bridge_ip,
                                          logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Configuring packetbeat on {container.docker_gw_bridge_ip}, {container.get_full_name()}, "
                        f"ips: {container.get_ips()}")
            csle_collector.host_manager.query_host_manager.config_packetbeat(
                stub=stub, kibana_ip=emulation_env_config.elk_config.container.get_ips()[0],
                kibana_port=emulation_env_config.elk_config.kibana_port,
                elastic_ip=emulation_env_config.elk_config.container.get_ips()[0],
                elastic_port=emulation_env_config.elk_config.elastic_port,
                num_elastic_shards=emulation_env_config.beats_config.num_elastic_shards)

    @staticmethod
    def config_metricbeat(emulation_env_config: EmulationEnvConfig, container: NodeContainerConfig,
                          logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container
        to setup the metricbeat configuration

        :param emulation_env_config: the emulation env config
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=container.docker_gw_bridge_ip,
                                          logger=logger)
        node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=container.get_ips())
        if node_beats_config is None:
            return

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Configuring metricbeat on {container.docker_gw_bridge_ip}, {container.get_full_name()}, "
                        f"ips: {container.get_ips()}")
            csle_collector.host_manager.query_host_manager.config_metricbeat(
                stub=stub, kibana_ip=emulation_env_config.elk_config.container.get_ips()[0],
                kibana_port=emulation_env_config.elk_config.kibana_port,
                elastic_ip=emulation_env_config.elk_config.container.get_ips()[0],
                elastic_port=emulation_env_config.elk_config.elastic_port,
                num_elastic_shards=emulation_env_config.beats_config.num_elastic_shards,
                kafka_ip=emulation_env_config.kafka_config.container.get_ips()[0],
                kafka_port=emulation_env_config.kafka_config.kafka_port,
                metricbeat_modules=node_beats_config.metricbeat_modules,
                reload_enabled=emulation_env_config.beats_config.reload_enabled)

    @staticmethod
    def config_heartbeat(emulation_env_config: EmulationEnvConfig, container: NodeContainerConfig,
                         logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container
        to setup the heartbeat configuration

        :param emulation_env_config: the emulation env config
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=container.docker_gw_bridge_ip,
                                          logger=logger)
        node_beats_config = emulation_env_config.beats_config.get_node_beats_config_by_ips(ips=container.get_ips())
        if node_beats_config is None:
            return

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{container.docker_gw_bridge_ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Configuring heartbeat on {container.docker_gw_bridge_ip}, {container.get_full_name()}, "
                        f"ips: {container.get_ips()}")
            csle_collector.host_manager.query_host_manager.config_heartbeat(
                stub=stub, kibana_ip=emulation_env_config.elk_config.container.get_ips()[0],
                kibana_port=emulation_env_config.elk_config.kibana_port,
                elastic_ip=emulation_env_config.elk_config.container.get_ips()[0],
                elastic_port=emulation_env_config.elk_config.elastic_port,
                num_elastic_shards=emulation_env_config.beats_config.num_elastic_shards,
                hosts_to_monitor=node_beats_config.heartbeat_hosts_to_monitor)

    @staticmethod
    def stop_host_monitor_threads(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                                  physical_host_ip: str) -> None:
        """
        A method that sends a request to the HostManager on every container to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param physical_host_ip: the IP of the physical host
        :param logger: the logger to use for logging
        :return: None
        """
        # Stop host monitor threads on emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_host_ip:
                HostController.stop_host_monitor_thread(emulation_env_config=emulation_env_config,
                                                        ip=c.docker_gw_bridge_ip, logger=logger)
        if emulation_env_config.kafka_config.container.physical_host_ip == physical_host_ip:
            # Stop host monitor threads on the kafka container
            HostController.stop_host_monitor_thread(emulation_env_config=emulation_env_config,
                                                    ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                                                    logger=logger)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_host_ip:
            # Stop host monitor threads on the ELK container
            HostController.stop_host_monitor_thread(emulation_env_config=emulation_env_config,
                                                    ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                                                    logger=logger)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_host_ip:
                # Stop host monitor threads on the SDN controller container
                HostController.stop_host_monitor_thread(
                    emulation_env_config=emulation_env_config,
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip, logger=logger)

    @staticmethod
    def stop_host_monitor_thread(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container to stop the monitor threads

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Stopping the Host monitor thread on {ip}.")
            csle_collector.host_manager.query_host_manager.stop_host_monitor(stub=stub)

    @staticmethod
    def stop_filebeat(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container to stop filebeat

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Stopping filebeat on {ip}.")
            csle_collector.host_manager.query_host_manager.stop_filebeat(stub=stub)

    @staticmethod
    def stop_packetbeat(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container to stop packetbeat

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Stopping packetbeat on {ip}.")
            csle_collector.host_manager.query_host_manager.stop_packetbeat(stub=stub)

    @staticmethod
    def stop_metricbeat(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container to stop metricbeat

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Stopping metricbeat on {ip}.")
            csle_collector.host_manager.query_host_manager.stop_metricbeat(stub=stub)

    @staticmethod
    def stop_heartbeat(emulation_env_config: EmulationEnvConfig, ip: str, logger: logging.Logger) -> None:
        """
        A method that sends a request to the HostManager on a specific container to stop heartbeat

        :param emulation_env_config: the emulation env config
        :param ip: the IP of the container
        :param logger: the logger to use for logging
        :return: None
        """
        HostController.start_host_manager(emulation_env_config=emulation_env_config, ip=ip, logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{ip}:'
                f'{emulation_env_config.host_manager_config.host_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            logger.info(f"Stopping heartbeat on {ip}.")
            csle_collector.host_manager.query_host_manager.stop_heartbeat(stub=stub)

    @staticmethod
    def get_host_monitor_threads_statuses(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                          logger: logging.Logger) \
            -> List[csle_collector.host_manager.host_manager_pb2.HostStatusDTO]:
        """
        A method that sends a request to the HostManager on every container to get the status of the Host monitor thread

        :param emulation_env_config: the emulation config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: List of monitor thread statuses
        """
        statuses = []
        HostController.start_host_managers(emulation_env_config=emulation_env_config, logger=logger)

        # Get statuses of emulation containers
        for c in emulation_env_config.containers_config.containers:
            if c.physical_host_ip == physical_server_ip:
                status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                    ip=c.docker_gw_bridge_ip, port=emulation_env_config.host_manager_config.host_manager_port)
                statuses.append(status)

        if emulation_env_config.kafka_config.container.physical_host_ip == physical_server_ip:
            # Get status of kafka container
            status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                ip=emulation_env_config.kafka_config.container.docker_gw_bridge_ip,
                port=emulation_env_config.host_manager_config.host_manager_port)
            statuses.append(status)

        if emulation_env_config.elk_config.container.physical_host_ip == physical_server_ip:
            # Get status of ELK container
            status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                ip=emulation_env_config.elk_config.container.docker_gw_bridge_ip,
                port=emulation_env_config.host_manager_config.host_manager_port)
            statuses.append(status)

        if emulation_env_config.sdn_controller_config is not None:
            if emulation_env_config.sdn_controller_config.container.physical_host_ip == physical_server_ip:
                # Get status of SDN controller container
                status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                    ip=emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip,
                    port=emulation_env_config.host_manager_config.host_manager_port)
                statuses.append(status)

        return statuses

    @staticmethod
    def get_host_monitor_thread_status_by_port_and_ip(ip: str, port: int,
                                                      timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) -> \
            csle_collector.host_manager.host_manager_pb2.HostStatusDTO:
        """
        A method that sends a request to the HostManager on a specific container
        to get the status of the Host monitor thread

        :param ip: the ip of the container
        :param port: the port of the host manager
        :param timeout: the timeout of the GRPC query
        :return: the status of the host manager
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.host_manager.host_manager_pb2_grpc.HostManagerStub(channel)
            status = csle_collector.host_manager.query_host_manager.get_host_status(stub=stub, timeout=timeout)
            return status

    @staticmethod
    def get_host_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the ips of the Host managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []

        # Get ips of emulation containers
        for c in emulation_env_config.containers_config.containers:
            ips.append(c.docker_gw_bridge_ip)

        # Get ip of Kafka container
        ips.append(emulation_env_config.kafka_config.container.docker_gw_bridge_ip)

        # Get ip of ELK container
        ips.append(emulation_env_config.elk_config.container.docker_gw_bridge_ip)

        if emulation_env_config.sdn_controller_config is not None:
            # Get ip of SDN controller container
            ips.append(emulation_env_config.sdn_controller_config.container.docker_gw_bridge_ip)

        return ips

    @staticmethod
    def get_host_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Host managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []

        # Get port of emulation containers
        for c in emulation_env_config.containers_config.containers:
            ports.append(emulation_env_config.host_manager_config.host_manager_port)

        # Get port of kafka container
        ports.append(emulation_env_config.host_manager_config.host_manager_port)

        # Get port of ELK container
        ports.append(emulation_env_config.host_manager_config.host_manager_port)

        if emulation_env_config.sdn_controller_config is not None:
            # Get port of SDN controller container
            ports.append(emulation_env_config.host_manager_config.host_manager_port)

        return ports

    @staticmethod
    def get_host_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                               logger: logging.Logger, physical_host_ip: str) -> HostManagersInfo:
        """
        Extracts the information of the Host managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param logger: the logger to use for logging
        :param physical_host_ip: the ip of the physical host
        :return: a DTO with the status of the Host managers
        """
        host_managers_ips = HostController.get_host_managers_ips(emulation_env_config=emulation_env_config)
        host_managers_ports = HostController.get_host_managers_ports(emulation_env_config=emulation_env_config)
        host_managers_statuses = []
        host_managers_running = []
        for ip in host_managers_ips:
            if ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=ip, physical_host_ip=physical_host_ip):
                continue
            status = None
            running = False
            try:
                status = HostController.get_host_monitor_thread_status_by_port_and_ip(
                    port=emulation_env_config.host_manager_config.host_manager_port, ip=ip)
                running = True
            except Exception as e:
                logger.info(f"Could not fetch Host manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                host_managers_statuses.append(status)
            else:
                host_managers_statuses.append(
                    csle_collector.host_manager.host_manager_util.HostManagerUtil.host_monitor_dto_empty())
            host_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        host_manager_info_dto = HostManagersInfo(host_managers_running=host_managers_running, ips=host_managers_ips,
                                                 execution_id=execution_id, emulation_name=emulation_name,
                                                 host_managers_statuses=host_managers_statuses,
                                                 ports=host_managers_ports)
        return host_manager_info_dto



from typing import List
import threading
import time
from csle_collector.host_manager.dao.host_metrics import HostMetrics
from csle_common.dao.emulation_observation.defender.emulation_defender_machine_observation_state import \
    EmulationDefenderMachineObservationState


class AggregatedHostMetricsThread(threading.Thread):
    """
    Thread that polls the Kafka log to get the latest status of the docker statistics for a specific host
    """

    def __init__(self, host_metrics: HostMetrics,
                 machines: List[EmulationDefenderMachineObservationState], sleep_time: int) -> None:
        """
        Initializes the thread

        :param host_metrics: the host metrics to update
        :param machines: the list of machines to update the host metrics with
        """
        threading.Thread.__init__(self)
        self.machines = machines
        self.running = True
        self.host_metrics = host_metrics
        self.host_metrics_list: List[HostMetrics] = []
        self.sleep_time = sleep_time

    def run(self) -> None:
        """
        Runs the thread

        :return: None
        """
        while self.running:
            time.sleep(self.sleep_time)
            total_num_logged_in_users = 0
            total_num_failed_login_attempts = 0
            total_num_open_connections = 0
            total_num_login_events = 0
            total_num_processes = 0
            total_num_users = 0
            for m in self.machines:
                if m.host_metrics is None:
                    raise ValueError("HostMetrics is None")
                total_num_logged_in_users += m.host_metrics.num_logged_in_users
                total_num_failed_login_attempts += m.host_metrics.num_failed_login_attempts
                total_num_open_connections += m.host_metrics.num_open_connections
                total_num_login_events += m.host_metrics.num_login_events
                total_num_processes += m.host_metrics.num_processes
                total_num_users += m.host_metrics.num_users
            self.host_metrics.num_logged_in_users = total_num_logged_in_users
            self.host_metrics.num_failed_login_attempts = total_num_failed_login_attempts
            self.host_metrics.num_open_connections = total_num_open_connections
            self.host_metrics.num_login_events = total_num_login_events
            self.host_metrics.num_processes = total_num_processes
            self.host_metrics.num_users = total_num_users
            self.host_metrics_list.append(self.host_metrics.copy())

    def get_average_aggregated_host_metrics(self) -> HostMetrics:
        """
        :return: average of the list of aggregated host metrics
        """
        if len(self.host_metrics_list) == 0:
            return self.host_metrics.copy()
        if len(self.host_metrics_list) == 1:
            return self.host_metrics_list[0].copy()
        avg_host_metrics = HostMetrics()
        for i in range(0, len(self.host_metrics_list)):
            avg_host_metrics.num_logged_in_users = (avg_host_metrics.num_logged_in_users +
                                                    self.host_metrics_list[i].num_logged_in_users)
            avg_host_metrics.num_failed_login_attempts = (avg_host_metrics.num_failed_login_attempts +
                                                          self.host_metrics_list[i].num_failed_login_attempts)
            avg_host_metrics.num_open_connections = (avg_host_metrics.num_open_connections
                                                     + self.host_metrics_list[i].num_open_connections)
            avg_host_metrics.num_login_events = (avg_host_metrics.num_login_events +
                                                 self.host_metrics_list[i].num_login_events)
            avg_host_metrics.num_processes = avg_host_metrics.num_processes + self.host_metrics_list[i].num_processes
            avg_host_metrics.num_users = avg_host_metrics.num_users + self.host_metrics_list[i].num_users

        avg_host_metrics.num_logged_in_users = int(round(avg_host_metrics.num_logged_in_users /
                                                         len(self.host_metrics_list)))
        avg_host_metrics.num_failed_login_attempts = int(round(avg_host_metrics.num_failed_login_attempts /
                                                               len(self.host_metrics_list)))
        avg_host_metrics.num_open_connections = int(round(avg_host_metrics.num_open_connections /
                                                          len(self.host_metrics_list)))
        avg_host_metrics.num_login_events = int(round(avg_host_metrics.num_login_events / len(self.host_metrics_list)))
        avg_host_metrics.num_processes = int(round(avg_host_metrics.num_processes / len(self.host_metrics_list)))
        avg_host_metrics.num_users = int(round(avg_host_metrics.num_users / len(self.host_metrics_list)))
        return avg_host_metrics



from typing import Dict, Any, Union
from csle_base.json_serializable import JSONSerializable


class ClusterNode(JSONSerializable):
    """
    A DTO Class representing a node in a CSLE cluster
    """

    def __init__(self, ip: str, leader: bool, cpus: int, gpus: int, RAM: int):
        """
        Initializes the DTO

        :param ip: the ip of the node
        :param leader: boolean indicating whether the node is the leader node or not
        :param cpus: the number of CPU cores of the node
        :param gpus: the number of GPUs of the node
        :param RAM: the amount of RAM of the node (GB)
        """
        self.ip = ip
        self.leader = leader
        self.cpus = cpus
        self.gpus = gpus
        self.RAM = RAM

    def to_dict(self) -> Dict[str, Union[str, bool, int]]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Union[str, bool, int]] = {}
        d["ip"] = self.ip
        d["leader"] = self.leader
        d["cpus"] = self.cpus
        d["gpus"] = self.gpus
        d["RAM"] = self.RAM
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "ClusterNode":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = ClusterNode(ip=d["ip"], leader=d["leader"], cpus=d["cpus"], gpus=d["gpus"], RAM=d["RAM"])
        return dto

    def __str__(self) -> str:
        """
        :return: a string representation of the credential
        """
        return f"ip: {self.ip}, leader: {self.leader}, cpus: {self.cpus}, gpus: {self.gpus}, RAM: {self.RAM}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "ClusterNode":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return ClusterNode.from_dict(json.loads(json_str))

    def copy(self) -> "ClusterNode":
        """
        :return: a copy of the DTO
        """
        return ClusterNode.from_dict(self.to_dict())

    @staticmethod
    def schema() -> "ClusterNode":
        """
        :return: get the schema of the DTO
        """
        return ClusterNode(ip="", leader=False, cpus=0, gpus=0, RAM=0)



from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.controllers.five_g_cu_controller import FiveGCUController
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.five_g_config import FiveGConfig
import pytest


class TestFiveGCUControllerSuite:
    """
    Test five_g_cu_controller
    """

    @pytest.fixture(autouse=True)
    def logger_setup(self) -> None:
        """
        Set up logger

        :return: None
        """
        self.logger = MagicMock()

    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.start_five_g_cu")
    def test_start_five_g_cus(self, mock_start_five_g_cu, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for starting the 5G CUs

        :param mock_start_five_g_cu: mock_start_five_g_cu
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES = \
            [example_emulation_env_config.containers_config.containers[0].name]
        FiveGCUController.start_five_g_cus(
            emulation_env_config=example_emulation_env_config,
            physical_server_ip=example_emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_start_five_g_cu.assert_called_once_with(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)

    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.stop_five_g_cu")
    def test_stop_five_g_cus(self, mock_stop_five_g_cu, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for stopping the 5G CUs

        :param mock_stop_five_g_cu: mock_stop_five_g_cu
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        constants.CONTAINER_IMAGES.FIVE_G_CU_IMAGES = \
            [example_emulation_env_config.containers_config.containers[0].name]
        FiveGCUController.stop_five_g_cus(
            emulation_env_config=example_emulation_env_config,
            physical_server_ip=example_emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        mock_stop_five_g_cu.assert_called_once_with(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)

    @patch("csle_collector.five_g_cu_manager.query_five_g_cu_manager.start_five_g_cu")
    @patch("grpc.insecure_channel")
    def test_start_five_g_cu(self, mock_insecure_channel, mock_start_five_g_cu,
                             example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for starting the 5G CU on a specific IP

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_start_five_g_cu: mock_start_five_g_cu
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        constants.GRPC_SERVERS.GRPC_OPTIONS = []
        mock_channel = MagicMock()
        mock_insecure_channel.return_value = mock_channel
        mock_stub = MagicMock()
        mock_channel.__enter__.return_value = mock_stub
        FiveGCUController.start_five_g_cu(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Starting the 5G CU on container with ip "
            f"{example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip} in "
            f"execution {example_emulation_env_config.execution_id} "
            f"of emulation: {example_emulation_env_config.name}")
        mock_insecure_channel.assert_called()
        mock_start_five_g_cu.assert_called()

    @patch("csle_collector.five_g_cu_manager.query_five_g_cu_manager.stop_five_g_cu")
    @patch("grpc.insecure_channel")
    def test_stop_five_g_cu(self, mock_insecure_channel, mock_stop_five_g_cu,
                            example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for stopping the 5G CU on a specific IP

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_stop_five_g_cu: mock_stop_five_g_cu
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        constants.GRPC_SERVERS.GRPC_OPTIONS = []
        mock_channel = MagicMock()
        mock_insecure_channel.return_value = mock_channel
        mock_stub = MagicMock()
        mock_channel.__enter__.return_value = mock_stub
        FiveGCUController.stop_five_g_cu(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip,
            logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Stopping the 5G CU on container with ip "
            f"{example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip} in "
            f"execution {example_emulation_env_config.execution_id} "
            f"of emulation: {example_emulation_env_config.name}")
        mock_insecure_channel.assert_called()
        mock_stop_five_g_cu.assert_called()

    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.start_five_g_cu_manager")
    def test_start_five_g_cu_managers(self, start_five_g_cu_manager,
                                      example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for starting 5G CU managers

        :param start_five_g_cu_manager: start_five_g_cu_manager
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        FiveGCUController.start_five_g_cu_managers(
            emulation_env_config=example_emulation_env_config,
            physical_server_ip=example_emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        start_five_g_cu_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_start_five_g_cu_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin,
                                     example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for starting the 5G CU manager on a specific IP

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param example_emulation_env_config: example_emulation_env_config
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_connection = MagicMock()
        example_emulation_env_config.get_connection = MagicMock()  # type: ignore[method-assign]
        example_emulation_env_config.get_connection.return_value = mock_connection
        mock_execute_ssh_cmd.side_effect = [
            (b"", b"", 0),  # Output for stopping old background job
            (b"", b"", 0),  # Output for starting the five_g_cu_manager
        ]
        FiveGCUController.start_five_g_cu_manager(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_any_call(
            f"Starting 5G CU manager on node "
            f"{example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}")

    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.stop_five_g_cu_manager")
    def test_stop_five_g_cu_managers(self, stop_five_g_cu_manager,
                                     example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for stopping 5G CU managers

        :param stop_five_g_cu_manager: stop_five_g_cu_manager
        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        FiveGCUController.stop_five_g_cu_managers(
            emulation_env_config=example_emulation_env_config,
            physical_server_ip=example_emulation_env_config.containers_config.containers[0].physical_host_ip,
            logger=self.logger)
        stop_five_g_cu_manager.assert_called()

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    @patch("time.sleep", return_value=None)
    def test_stop_five_g_cu_manager(self, mock_sleep, mock_execute_ssh_cmd, mock_connect_admin,
                                    example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test utility function for stopping the 5G CU manager on a specific IP

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :param example_emulation_env_config: example_emulation_env_config
        :param mock_sleep: mock_sleep
        :return: None
        """
        mock_connection = MagicMock()
        example_emulation_env_config.get_connection = MagicMock()  # type: ignore[method-assign]
        example_emulation_env_config.get_connection.return_value = mock_connection
        mock_execute_ssh_cmd.side_effect = [(b"", b"", 0)]
        FiveGCUController.stop_five_g_cu_manager(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip, logger=self.logger)
        mock_connect_admin.assert_called_once_with(
            emulation_env_config=example_emulation_env_config,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)
        mock_execute_ssh_cmd.assert_called()
        self.logger.info.assert_any_call(
            f"Stopping 5G CU manager on node "
            f"{example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip}")

    def test_get_five_g_cu_managers_ips(self, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test a method that extracts the IPs of the 5G Cu managers in a given emulation

        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        ips = FiveGCUController.get_five_g_cu_managers_ips(emulation_env_config=example_emulation_env_config)
        expected_ips = [example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip]
        assert ips == expected_ips

    def test_get_five_g_cu_managers_ports(self, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test a method that extracts the ports of the 5G CU managers in a given emulation

        :param example_emulation_env_config: example_emulation_env_config
        :return: None
        """
        ports = FiveGCUController.get_five_g_cu_managers_ports(emulation_env_config=example_emulation_env_config)
        expected_ports = [50053]
        assert ports == expected_ports

    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.get_five_g_cu_managers_ips")
    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController.get_five_g_cu_managers_ports")
    @patch("csle_common.controllers.five_g_cu_controller.FiveGCUController."
           "get_five_g_cu_status_by_ip_and_port")
    @patch("csle_common.util.emulation_util.EmulationUtil.physical_ip_match")
    def test_get_five_g_cu_managers_info(self, mock_physical_ip_match, mock_get_statuses, mock_get_ports,
                                         mock_get_ips, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Test the method that extracts the information of the 5G CU managers for a given emulation

        :param mock_physical_ip_match: mock_physical_ip_match
        :param mock_get_statuses: mock_get_statuses
        :param mock_get_ports:mock_get_ports
        :param mock_get_ips: mock_get_ips
        :return: None
        """
        mock_get_ips.return_value = [example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip]
        mock_get_ports.return_value = [example_emulation_env_config.five_g_config.five_g_cu_manager_port]
        mock_status = MagicMock()
        mock_get_statuses.side_effect = [mock_status, Exception("Test exception")]
        mock_physical_ip_match.side_effect = [True, False]
        active_ips = [example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip]
        physical_server_ip = example_emulation_env_config.containers_config.containers[0].physical_host_ip
        FiveGCUController.get_five_g_cu_managers_info(
            emulation_env_config=example_emulation_env_config, active_ips=active_ips, logger=self.logger,
            physical_server_ip=physical_server_ip)
        mock_get_ips.assert_called_once_with(emulation_env_config=example_emulation_env_config)
        mock_get_ports.assert_called_once_with(emulation_env_config=example_emulation_env_config)
        mock_get_statuses.assert_any_call(
            port=example_emulation_env_config.five_g_config.five_g_cu_manager_port,
            ip=example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip)

    @patch("csle_collector.five_g_cu_manager.query_five_g_cu_manager.init_five_g_cu")
    @patch("grpc.insecure_channel")
    def test_init_five_g_cu(self, mock_insecure_channel, mock_init_five_g_cu,
                            example_emulation_env_config: EmulationEnvConfig,
                            example_containers_config_five_g: NodeContainerConfig,
                            example_five_g_config_two: FiveGConfig) -> None:
        """
        Test utility function for initializing the 5G UE on a specific IP

        :param mock_insecure_channel: mock_insecure_channel
        :param mock_init_five_g_cu: mock_init_five_g_cu
        :param example_emulation_env_config: example_emulation_env_config
        :param example_containers_config_five_g: example_containers_config_five_g
        :param example_five_g_config_two: example_five_g_config_two
        :return: None
        """
        constants.GRPC_SERVERS.GRPC_OPTIONS = []
        mock_channel = MagicMock()
        mock_insecure_channel.return_value = mock_channel
        mock_stub = MagicMock()
        mock_channel.__enter__.return_value = mock_stub
        result = FiveGCUController.init_five_g_cu(
            emulation_env_config=example_emulation_env_config,
            container=example_emulation_env_config.containers_config.containers[0],
            logger=self.logger)
        self.logger.info.assert_called_once_with(
            f"Initializing the 5G CU on container with ip "
            f"{example_emulation_env_config.containers_config.containers[0].docker_gw_bridge_ip} "
            f"in execution {example_emulation_env_config.execution_id} "
            f"of emulation: {example_emulation_env_config.name}")
        assert result is None
        example_emulation_env_config.containers_config = example_containers_config_five_g
        example_emulation_env_config.five_g_config = example_five_g_config_two
        result = FiveGCUController.init_five_g_cu(
            emulation_env_config=example_emulation_env_config,
            container=example_emulation_env_config.containers_config.containers[4],
            logger=self.logger)
        assert result is not None
        mock_insecure_channel.assert_called()
        mock_init_five_g_cu.assert_called()



from enum import IntEnum


class AgentType(IntEnum):
    """
    Enum representing the different agent types in CSLE
    """
    T_SPSA = 0
    PPO = 1
    T_FP = 2
    DQN = 3
    REINFORCE = 4
    NFSP = 5
    RANDOM = 6
    NONE = 7
    VALUE_ITERATION = 8
    HSVI = 9
    SONDIK_VALUE_ITERATION = 10
    RANDOM_SEARCH = 11
    DIFFERENTIAL_EVOLUTION = 12
    CROSS_ENTROPY = 13
    KIEFER_WOLFOWITZ = 14
    Q_LEARNING = 15
    SARSA = 16
    POLICY_ITERATION = 17
    SHAPLEY_ITERATION = 18
    HSVI_OS_POSG = 19
    FICTITIOUS_PLAY = 20
    LINEAR_PROGRAMMING_NORMAL_FORM = 21
    DYNA_SEC = 22
    BAYESIAN_OPTIMIZATION = 23
    DFSP_LOCAL = 24
    LINEAR_PROGRAMMING_CMDP = 25
    BAYESIAN_OPTIMIZATION_EMUKIT = 26
    SIMULATED_ANNEALING = 27
    CMA_ES = 28
    NELDER_MEAD = 29
    PARTICLE_SWARM = 30
    PPO_CLEAN = 31
    POMCP = 32
    DQN_CLEAN = 33
    C51_CLEAN = 34
    PPG_CLEAN = 35
    MCS = 36



from csle_common.util.plotting_util import PlottingUtil
from scipy import stats
import numpy as np


class TestPlottingUtilSuite:
    """
    Test suite for plotting util
    """

    def test_running_average(self) -> None:
        """
        Test the function used to compute the running average of the last N elements of a vector x

        :return: None
        """
        x = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        N = 3
        expected = np.array([1, 2, 3, 3, 4, 5, 6, 7, 8, 9])
        result = PlottingUtil.running_average(x, N)
        assert result.any() == expected.any()

    def test_mean_confidence_interval(self) -> None:
        """
        Test function that computes confidence intervals

        :return: None
        """
        data = np.array([1, 2, 3, 4, 5])
        mean, h = PlottingUtil.mean_confidence_interval(data=data, confidence=0.95)
        expected_mean = np.mean(data)
        expected_se = stats.sem(data)
        expected_h = expected_se * stats.t.ppf((1 + 0.95) / 2.0, len(data) - 1)
        assert expected_mean == mean
        assert expected_h == h

    def test_min_max_norm(self) -> None:
        """
        Test function that computes min-max normalization of a vector

        :return: None
        """
        vec = np.array([1, 2, 3, 4, 5])
        min_val = 1
        max_val = 5
        expected = np.array([0.0, 0.25, 0.5, 0.75, 1.0])
        result = PlottingUtil.min_max_norm(vec, max_val, min_val)
        assert result.any() == expected.any()



from typing import Optional, List, Tuple, Dict, Any
import time
import csle_common.constants.constants as constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_type import EmulationAttackerActionType
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_outcome import EmulationAttackerActionOutcome
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class EmulationAttackerAction(JSONSerializable):
    """
    Class representing an action of the attacker in the emulation
    """
    def __init__(self, id: EmulationAttackerActionId, name: str, cmds: List[str],
                 type: EmulationAttackerActionType, descr: str, ips: List[str], index: int,
                 action_outcome: EmulationAttackerActionOutcome = EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                 vulnerability: Optional[str] = None, alt_cmds: Optional[List[str]] = None, backdoor: bool = False,
                 execution_time: float = 0.0, ts: Optional[float] = None):
        """
        Class constructor

        :param id: id of the action
        :param name: name of the action
        :param cmds: command-line commands to apply the action on the emulation
        :param type: type of the action
        :param descr: description of the action (documentation)
        :param ips: ips of the machines to apply the action to
        :param index: index of the machien to apply the action to
        :param action_outcome: type of the outcome of the action
        :param vulnerability: type of vulnerability that the action exploits (in case an exploit)
        :param alt_cmds: alternative command if the first command does not work
        :param backdoor: if the action also installs a backdoor (some exploits does this)
        :param execution_time: the time it took to run the action
        :param ts: the timestep the action was completed
        """
        self.type = type
        self.id = id
        self.name = name
        self.cmds = cmds
        self.descr = descr
        self.index = index
        self.ips = ips
        self.vulnerability = vulnerability
        self.action_outcome = action_outcome
        self.backdoor = backdoor
        self.alt_cmds = alt_cmds
        if self.alt_cmds is None:
            self.alt_cmds = []
        self.index = index
        self.backdoor = backdoor
        self.execution_time = round(execution_time, 3)
        self.ts = ts

    def nmap_cmds(self, machine_ips: Optional[List[str]] = None) -> Tuple[List[str], List[str]]:
        """
        Augments the original command of the action with extra flags for NMAP

        :param machine_ips: list of ips
        :return: the new command
        """
        commands = []
        file_names = []
        for ip in self.ips:
            file_name = str(self.id) + "_" + str(self.index) + "_" + ip.replace("/", "_")
            file_name = file_name + ".xml "
            commands.append(self.cmds[0] + constants.NMAP.FILE_ARGS + " " + file_name + ip)
            file_names.append(file_name)
        if machine_ips is None:
            machine_ips = []
        for ip in machine_ips:
            file_name = str(self.id) + "_" + str(self.index) + "_" + ip.replace("/", "_")
            file_name = file_name + ".xml "
            commands.append(self.cmds[0] + constants.NMAP.FILE_ARGS + " " + file_name + ip)
            file_names.append(file_name)
        return commands, file_names

    def nikto_cmds(self) -> Tuple[List[str], List[str]]:
        """
        Augments the original command of the action with extra flags for Nikto

        :return: the new command
        """
        file_names = []
        commands = []
        for ip in self.ips:
            file_name = str(self.id) + "_" + str(self.index) + "_" + ip + ".xml "
            commands.append(self.cmds[0] + constants.NIKTO.HOST_ARG + ip + " " + constants.NIKTO.OUTPUT_ARG + file_name)
            file_names.append(file_name)
        return commands, file_names

    def masscan_cmds(self) -> Tuple[List[str], List[str]]:
        """
        Augments the original command of the action with extra flags for massscan

        :return: the new command
        """
        commands = []
        file_names = []
        for ip in self.ips:
            file_name = str(self.id) + "_" + str(self.index) + "_" + ip + ".xml "
            if self.index == -1:
                file_name = str(self.id) + "_" + str(self.index) + ".xml "
            commands.append(self.cmds[0] + constants.MASSCAN.OUTPUT_ARG + " " + file_name + ip)
            file_names.append(file_name)
        return commands, file_names

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return "id:{},name:{},ips:{},index:{}".format(self.id, self.name, self.ips, self.index)

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationAttackerAction":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationAttackerAction(
            type=d["type"], id=d["id"], name=d["name"], cmds=d["cmds"], descr=d["descr"], index=d["index"],
            ips=d["ips"], vulnerability=d["vulnerability"], action_outcome=d["action_outcome"],
            backdoor=d["backdoor"], alt_cmds=d["alt_cmds"],
            execution_time=d["execution_time"], ts=d["ts"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["name"] = self.name
        d["cmds"] = list(self.cmds)
        d["type"] = self.type
        d["descr"] = self.descr
        d["ips"] = list(self.ips)
        d["index"] = self.index
        d["action_outcome"] = self.action_outcome
        d["vulnerability"] = self.vulnerability
        d["alt_cmds"] = list(self.alt_cmds) if self.alt_cmds is not None else self.alt_cmds
        d["backdoor"] = self.backdoor
        d["execution_time"] = self.execution_time
        d["ts"] = self.ts
        return d

    def ips_match(self, ips: List[str]) -> bool:
        """
        Checks if a list of ips overlap with the ips of this host

        :param ips: the list of ips to check
        :return:  True if they match, False otherwise
        """
        for ip in self.ips:
            if ip in ips:
                return True
        return False

    def to_kafka_record(self) -> str:
        """
        Converts the instance into a kafka record format

        :param total_time: the total time of execution
        :return: the kafka record
        """
        if self.vulnerability is None:
            vuln = ""
        else:
            vuln = self.vulnerability
        if self.alt_cmds is None:
            alt_cmds = []
        else:
            alt_cmds = self.alt_cmds
        ts = time.time()
        if self.alt_cmds is None or len(self.alt_cmds) == 0:
            self.alt_cmds = ["-"]
        if self.cmds is None or len(self.cmds) == 0:
            self.cmds = ["-"]
        alt_cmds = list(map(lambda x: x.replace(",", "-"), alt_cmds))
        vuln = vuln.replace(",", "-")
        cmds = list(map(lambda x: x.replace(",", "-"), self.cmds))
        record = f"{ts},{self.id},{self.descr.replace(',', '')},{self.index},{self.name.replace(',', '-')}," \
                 f"{self.execution_time},{'_'.join(self.ips)},{'_'.join(cmds)},{self.type}," \
                 f"{self.action_outcome},{vuln}," \
                 f"{'_'.join(alt_cmds)},{self.backdoor}"
        return record

    @staticmethod
    def from_kafka_record(record: str) -> "EmulationAttackerAction":
        """
        Converts a kafka record into an instance

        :param record: the record to convert
        :return: the created instance
        """
        parts = record.split(",")
        obj = EmulationAttackerAction(id=EmulationAttackerActionId(int(parts[1])), ts=float(parts[0]),
                                      descr=parts[2], index=int(parts[3]),
                                      name=parts[4],
                                      execution_time=float(parts[5]), ips=parts[6].split("_"),
                                      cmds=parts[7].split("_"),
                                      type=EmulationAttackerActionType(int(parts[8])),
                                      action_outcome=EmulationAttackerActionOutcome(int(parts[9])),
                                      vulnerability=parts[10], alt_cmds=parts[11].split("_"),
                                      backdoor=parts[12] == "True")
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationAttackerAction":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationAttackerAction.from_dict(json.loads(json_str))

    def copy(self) -> "EmulationAttackerAction":
        """
        :return: a copy of the DTO
        """
        return EmulationAttackerAction.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "EmulationAttackerAction":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.ips = list(map(lambda x: GeneralUtil.replace_first_octet_of_ip(ip=x, ip_first_octet=ip_first_octet),
                              config.ips))
        return config

    def num_attributes(self) -> int:
        """
        :return: The number of attributes of the DTO
        """
        return 13

    @staticmethod
    def schema() -> "EmulationAttackerAction":
        """
        :return: get the schema of the DTO
        """
        return EmulationAttackerAction(id=EmulationAttackerActionId.STOP, name="", cmds=[""],
                                       type=EmulationAttackerActionType.STOP, descr="", ips=[""], index=-1,
                                       action_outcome=EmulationAttackerActionOutcome.INFORMATION_GATHERING,
                                       vulnerability="", alt_cmds=[""], backdoor=False, execution_time=0.0, ts=0.0)



from typing import Tuple
import torch
import torch.nn as nn
from csle_common.dao.training.agent_type import AgentType


class QNetwork(nn.Module):
    """
    Class for instantiating a neural network for DQN training
    """

    def __init__(self, input_dim: int, num_hidden_layers: int, hidden_layer_dim: int,
                 agent_type: int, action_space_dim: int = 1,
                 n_atoms: int = 101, start: int = -100, end: int = 100, steps: int = 101) -> None:
        """
        Initializes the neural network

        :param input_dim: the input layer dimension
        :param num_hidden_layers: the number of hidden layers
        :param hidden_layer_dim: the dimension of each hidden layer; in agents referred to as num_hl_neur
        :param agent_type: the type of the agent
        :param action_space_dim:the action space dimension
        :param n_atoms: the number of atoms (for C51)
        :param start: the start of the atoms
        :param end: the end of the atoms
        :param steps: the number of steps for the atoms
        """
        super(QNetwork, self).__init__()

        self.num_hidden_layers = num_hidden_layers
        self.input_dim = input_dim
        self.agent_type = agent_type
        self.hidden_layer_dim = hidden_layer_dim
        self.action_space_dim = action_space_dim

        if self.agent_type == AgentType.DQN_CLEAN or agent_type == AgentType.DQN:
            self.network = nn.Sequential()
            for layer in range(self.num_hidden_layers):
                self.network.add_module(name=f'Layer {layer}', module=nn.Linear(input_dim, self.hidden_layer_dim))
                self.network.add_module(name='activation', module=nn.ReLU())
                input_dim = self.hidden_layer_dim
        elif self.agent_type == AgentType.C51_CLEAN:
            self.start = start
            self.end = end
            self.steps = steps
            self.n_atoms = n_atoms
            self.network = nn.Sequential()
            for layer in range(self.num_hidden_layers):
                self.network.add_module(name=f'Layer {layer}',
                                        module=nn.Linear(input_dim, self.hidden_layer_dim * n_atoms))
                self.network.add_module(name='activation', module=nn.ReLU())
                input_dim = self.hidden_layer_dim * self.n_atoms
            self.network.add_module(name='Output layer', module=nn.Linear(input_dim, self.action_space_dim * n_atoms))
        else:
            raise ValueError(f"Agent type: {agent_type} not recognized")

    def get_action(self, x) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Gets the action and the probability distribution over actions

        :param x: the input observation(s)
        :return: the action(s)
        """
        logits = self.network(x)
        pmfs = torch.softmax(logits.view(len(x), self.action_space_dim, self.n_atoms), dim=2)
        self.atoms = torch.linspace(self.start, self.end, self.steps)
        q_values = (pmfs * self.atoms).sum(2)
        action = torch.argmax(q_values, 1)
        return action, pmfs[torch.arange(len(x)), action]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Makes a forward pass of the neural network with a given input vector x

        :param x: the input vector
        :return: the output of the neural network
        """
        output: torch.Tensor = self.network(x)
        return output

    def save(self, path: str) -> None:
        """
        Saves the model to disk

        :param path: the path on disk to save the model
        :return: None
        """
        state_dict = self.state_dict()
        state_dict["input_dim"] = self.input_dim
        state_dict["num_hidden_layers"] = self.num_hidden_layers
        state_dict["hidden_layer_dim"] = self.hidden_layer_dim
        state_dict["agent_type"] = self.agent_type
        torch.save(state_dict, path)

    @staticmethod
    def load(path: str) -> "QNetwork":
        """
        Loads the model from a given path

        :param path: the path to load the model from
        :return: None
        """
        state_dict = torch.load(path)
        model = QNetwork(input_dim=state_dict["input_dim"], num_hidden_layers=state_dict["num_hidden_layers"],
                         hidden_layer_dim=state_dict["hidden_layer_dim"], agent_type=state_dict["agent_type"])
        del state_dict["input_dim"]
        del state_dict["num_hidden_layers"]
        del state_dict["hidden_layer_dim"]
        del state_dict["agent_type"]
        model.load_state_dict(state_dict)
        model.eval()
        return model



from unittest.mock import patch, MagicMock
from csle_common.util.export_util import ExportUtil
import os
import tempfile
import json


class TestExportUtilSuite:
    """
    Test suite for Export Util
    """

    def test_zipdir(self) -> None:
        """
        Test the method that creates a zip file of a given directory

        :return: None
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            file_path = os.path.join(tmpdir, "test.txt")
            with open(file_path, "w") as f:
                f.write("text")
            zip_path = os.path.join(tmpdir, "test.zip")
            ExportUtil.zipdir(tmpdir, zip_path)
            assert os.path.exists(zip_path)

    def test_get_dir_size_gb(self) -> None:
        """
        Test the method that calculates the zie of a file directory in gb

        :return: None
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            sub_dir = os.path.join(tmpdir, "subdir")
            os.makedirs(sub_dir)
            file_sizes = [150000000, 250000000, 350000000]
            total_size = sum(file_sizes)
            file_paths = [
                os.path.join(tmpdir, "file1.txt"),
                os.path.join(tmpdir, "file2.txt"),
                os.path.join(tmpdir, "file3.txt"),
            ]
            for file_path, size in zip(file_paths, file_sizes):
                with open(file_path, "wb") as f:
                    f.write(b"\0" * size)
            size_gb = ExportUtil.get_dir_size_gb(tmpdir)
            expected_size_gb = round(float(total_size) / 1000000000, 2)
            assert size_gb == expected_size_gb, f"Expected {expected_size_gb} GB but got {size_gb} GB"

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.list_emulation_traces_ids")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.get_emulation_trace")
    @patch("csle_common.util.export_util.ExportUtil.zipdir")
    def test_export_emulation_traces_to_disk_json(self, mock_zipdir,
                                                  mock_get_emulation_trace, mock_list_emulation_traces_ids) -> None:
        """
        Test the method that exports emulation traces from the metastore to disk

        :param mock_zipdir: mock_zipdir
        :param mock_get_emulation_trace: mock_get_emulation_trace
        :param mock_list_emulation_traces_ids: mock_list_emulation_traces_ids
        :return: None
        """
        mock_list_emulation_traces_ids.return_value = [(1,), (2,)]
        mock_trace = MagicMock()
        mock_trace.to_dict.return_value = {"trace_data": "example_data"}
        mock_trace.num_attributes_per_time_step.return_value = 5
        mock_trace.schema.return_value.to_dict.return_value = {"schema_data": "example_schema"}
        mock_get_emulation_trace.return_value = mock_trace
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_output = os.path.join(temp_dir, "output.zip")
            ExportUtil.export_emulation_traces_to_disk_json(
                num_traces_per_file=1,
                output_dir=temp_dir,
                zip_file_output=zip_file_output,
                max_num_traces=2,
                added_by="test_user",
                offset=0,
                file_start_id=1,
            )
            assert os.path.exists(os.path.join(temp_dir, "1.json"))

    @patch("csle_common.util.export_util.ExportUtil.get_dir_size_gb")
    @patch("os.listdir")
    @patch("os.path.getsize")
    @patch("io.open")
    def test_extract_emulation_traces_dataset_metadata(self, mock_open, mock_getsize, mock_listdir,
                                                       mock_get_dir_size_gb) -> None:
        """
        Test the method that extracts metadata of a traces dataset stored on disk

        :param mock_open: mock_open
        :param mock_getsize: mock_getsize
        :param mock_listdir: mock_listdir
        :param mock_get_dir_size_gb: mock_get_dir_size_gb
        :return: None
        """
        mock_metadata = {
            "schema": {"schema_data": "example_schema"},
            "num_traces": 5,
            "num_traces_per_file": 2,
            "num_attributes_per_time_step": 3,
            "file_format": "json",
            "added_by": "test_user",
            "columns": "example_columns",
        }
        mock_open.return_value.__enter__.return_value.read.return_value = json.dumps(mock_metadata)
        mock_listdir.return_value = ["1.json", "2.json", "metadata.json"]
        mock_getsize.return_value = 500000000
        mock_get_dir_size_gb.return_value = 1.5
        dir_path = "/dir"
        zip_file_path = "/dir/zipfile.zip"
        result = ExportUtil.extract_emulation_traces_dataset_metadata(dir_path=dir_path, zip_file_path=zip_file_path)
        assert result

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.list_emulation_traces_ids")
    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.get_emulation_trace")
    @patch("csle_common.util.export_util.ExportUtil.zipdir")
    def test_export_emulation_traces_to_disk_csv(self, mock_zipdir, mock_get_emulation_trace,
                                                 mock_list_emulation_traces_ids) -> None:
        """
        Test the method that exports emulation traces from the metastore to disk

        :param mock_zipdir: mock_zipdir
        :param mock_get_emulation_trace: mock_get_emulation_trace
        :param mock_list_emulation_traces_ids: mock_list_emulation_traces_id
        :return: None
        """
        mock_list_emulation_traces_ids.return_value = [(1,), (2,)]
        mock_trace = MagicMock()
        mock_trace.to_csv_record.return_value = (["value1", "value2"], ["col1", "col2"])
        mock_trace.num_attributes_per_time_step.return_value = 5
        mock_get_emulation_trace.return_value = mock_trace
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_output = os.path.join(temp_dir, "output.zip")
            ExportUtil.export_emulation_traces_to_disk_csv(
                num_traces_per_file=1,
                output_dir=temp_dir,
                zip_file_output=zip_file_output,
                max_num_traces=2,
                max_time_steps=10,
                max_nodes=5,
                max_ports=3,
                max_vulns=2,
                null_value=-1,
                added_by="test_user",
            )
            assert os.path.exists(os.path.join(temp_dir, "1.csv"))

    @patch("csle_common.metastore.metastore_facade.MetastoreFacade.get_emulation_statistic")
    @patch("csle_common.util.export_util.ExportUtil.zipdir")
    def test_export_emulation_statistic_to_disk_json(self, mock_zipdir, mock_get_emulation_statistic) -> None:
        """
        Test the method that exports emulation statistics from the metastore to disk

        :param mock_zipdir: mock_zipdir
        :param mock_get_emulation_statistic: mock_get_emulation_statistic
        :return: None
        """
        mock_statistic = MagicMock()
        mock_statistic.num_measurements = 100
        mock_statistic.num_metrics = 10
        mock_statistic.metrics = ["metric1", "metric2"]
        mock_statistic.conditions = ["condition1", "condition2"]
        mock_statistic.num_conditions = 2
        mock_statistic.to_dict.return_value = {"key": "value"}
        mock_get_emulation_statistic.return_value = mock_statistic
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_output = os.path.join(temp_dir, "output.zip")
            ExportUtil.export_emulation_statistics_to_disk_json(
                output_dir=temp_dir, zip_file_output=zip_file_output, statistics_id=1, added_by="test_user"
            )
            assert os.path.exists(os.path.join(temp_dir, "statistics.json"))

    @patch("csle_common.util.export_util.ExportUtil.get_dir_size_gb")
    @patch("os.listdir")
    @patch("os.path.getsize")
    @patch("io.open")
    def test_extract_emulation_statistics_dataset_metadata(self, mock_open, mock_getsize, mock_listdir,
                                                           mock_get_dir_size_gb) -> None:
        """
        Test the method that extracts metadata of a traces dataset stored on disk

        :param mock_open: mock_open
        :param mock_getsize: mock_getsize
        :param mock_listdir: mock_listdir
        :param mock_get_dir_size_gb: mock_get_dir_size_gb
        :return: None
        """
        mock_metadata = {
            "file_format": "json",
            "added_by": "test_user",
            "num_measurements": 100,
            "num_metrics": 10,
            "metrics": "metric1,metric2",
            "conditions": "condition1,condition2",
            "num_conditions": 2,
        }
        mock_open.return_value.__enter__.return_value.read.return_value = json.dumps(mock_metadata)
        mock_listdir.return_value = ["statistics.json", "metadata.json"]
        mock_getsize.return_value = 500000000
        mock_get_dir_size_gb.return_value = 1.5
        dir_path = "/dir"
        zip_file_path = "/dir/zipfile.zip"
        result = ExportUtil.extract_emulation_statistics_dataset_metadata(dir_path=dir_path,
                                                                          zip_file_path=zip_file_path)
        assert result



from typing import Union, List, Dict, Any
from csle_base.json_serializable import JSONSerializable


class HParam(JSONSerializable):
    """
    DTO class representing a hyperparameter
    """

    def __init__(self, value: Union[int, float, str, List[Any]], name: str, descr: str):
        """
        Initializes the DTO

        :param value: the value of the hyperparameter
        :param name: the name of the hyperparameter
        :param descr: the description of the hyperparameter
        """
        self.value = value
        self.name = name
        self.descr = descr

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d = {}
        d["value"] = self.value
        d["name"] = self.name
        d["descr"] = self.descr
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "HParam":
        """
        Creates an instance from a dict representation

        :param d: the dict reppresentation
        :return: the instance
        """
        obj = HParam(value=d["value"], name=d["name"], descr=d["descr"])
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "HParam":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return HParam.from_dict(json.loads(json_str))



from typing import List, Dict, Union, Optional, Any
import iteround
import numpy as np
from numpy.typing import NDArray
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.linear_threshold_stopping_policy import LinearThresholdStoppingPolicy
from csle_common.dao.training.tabular_policy import TabularPolicy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType


class LinearTabularPolicy(Policy):
    """
    A linear tabular policy that uses a linear threshold line to decide when to take action and a tabular policy to
    decide which action to take
    """

    def __init__(self, stopping_policy: LinearThresholdStoppingPolicy, action_policy: TabularPolicy,
                 simulation_name: str, states: List[State], player_type: PlayerType,
                 actions: List[Action], experiment_config: Optional[ExperimentConfig], avg_R: float,
                 agent_type: AgentType) -> None:
        """
        Initializes the policy

        :param simulation_name: the simulation name
        :param attacker: whether it is an attacker or not
        :param L: the number of stop actions
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration used for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        :param agent_type: the agent type
        :param opponent_strategy: optionally an opponent strategy
        """
        super(LinearTabularPolicy, self).__init__(agent_type=agent_type, player_type=player_type)
        self.stopping_policy = stopping_policy
        self.action_policy = action_policy
        self.id = -1
        self.simulation_name = simulation_name
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.policy_type = PolicyType.LINEAR_TABULAR

    def action(self, o: List[float], deterministic: bool = True) -> Union[int, List[int], float, NDArray[Any]]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        stop = self.stopping_policy.action(o=o[1:])
        if stop == 1:
            return self.action_policy.action(o=int(o[0]))
        else:
            return 0

    def probability(self, o: List[float], a: int) -> int:
        """
        Probability of a given action

        :param o: the current observation
        :param a: a given action
        :return: the probability of a
        """
        taken_action = self.action(o=o)
        return taken_action == a

    def stage_policy(self, o: Any) -> Any:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        stage_strategy = np.zeros((len(self.states), len(self.actions)))
        for i, s_a in enumerate(self.states):
            o[0] = s_a
            for j, a in enumerate(self.actions):
                stage_strategy[i][j] = self.probability(o=o, a=j)
            stage_strategy[i] = iteround.saferound(stage_strategy[i], 2)
            assert round(sum(stage_strategy[i]), 3) == 1
        return stage_strategy.tolist()

    def to_dict(self) -> Dict[str, List[float]]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["stopping_policy"] = self.stopping_policy.to_dict()
        d["action_policy"] = self.action_policy.to_dict()
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["player_type"] = self.player_type
        d["agent_type"] = self.agent_type
        if self.experiment_config is not None:
            d["experiment_config"] = self.experiment_config.to_dict()
        else:
            d["experiment_config"] = None
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "LinearTabularPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = LinearTabularPolicy(
            stopping_policy=LinearThresholdStoppingPolicy.from_dict(d["stopping_policy"]),
            action_policy=TabularPolicy.from_dict(d["action_policy"]),
            simulation_name=d["simulation_name"],
            states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
            actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
            experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"],
            agent_type=d["agent_type"])
        obj.id = d["id"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"stopping_policy: {self.stopping_policy}, action_policy: {self.action_policy}, " \
               f"id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"player_type: {self.player_type}, " \
               f"states: {self.states}, agent_type: {self.agent_type}, actions: {self.actions}," \
               f"experiment_config: {self.experiment_config}, avg_R: {self.avg_R}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "LinearTabularPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return LinearTabularPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "LinearTabularPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())



from typing import List, Dict, Any
from csle_common.dao.simulation_config.observation_space_config import ObservationSpaceConfig
from csle_base.json_serializable import JSONSerializable


class JointObservationSpaceConfig(JSONSerializable):
    """
    DTO representing a joint observation space configuration
    """

    def __init__(self, observation_spaces: List[ObservationSpaceConfig]):
        """
        Initializes the DTO

        :param observation_spaces: list of observation spaces
        """
        self.observation_spaces = observation_spaces

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "JointObservationSpaceConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = JointObservationSpaceConfig(
            observation_spaces=list(map(lambda x: ObservationSpaceConfig.from_dict(x), d["observation_spaces"]))
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d = {}
        d["observation_spaces"] = list(map(lambda x: x.to_dict(), self.observation_spaces))
        return d

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"observation spaces: {list(map(lambda x: str(x), self.observation_spaces))}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "JointObservationSpaceConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return JointObservationSpaceConfig.from_dict(json.loads(json_str))
